use rustix::{
    fd::AsFd,
    io::{OwnedFd, PollFd, PollFlags},
    net::{AddressFamily, Protocol, SocketFlags, SocketType},
};
use std::{collections::HashMap, ops::BitOr, sync::Arc};

pub mod net;

pub struct Key(usize);

pub struct PollManager {
    id: usize,
    notify: Notify,
    notify_token: NotifyToken,
    notify_key: Option<usize>,
    io: HashMap<usize, Managed>,
}

struct Managed {
    io: Io,
    interests: Readiness,
}

struct Io {
    inner: Arc<dyn AsFd>,
}

#[derive(Default)]
pub struct Poller<'a> {
    timeout: Option<i32>,
    flags: Vec<PollFlags>,
    fds: Vec<PollFd<'a>>,
}

#[derive(Clone, Copy, Default, PartialEq, Eq, Hash)]
pub struct Readiness {
    read: bool,
    write: bool,
    hangup: bool,
    error: bool,
}

pub struct Notify {
    registered: OwnedFd,
}

#[derive(Clone)]
pub struct NotifyToken {
    free: Arc<OwnedFd>,
}

pub fn notify_pair() -> rustix::io::Result<(Notify, NotifyToken)> {
    let (registered, free) = rustix::net::socketpair(
        AddressFamily::UNIX,
        SocketType::SEQPACKET,
        SocketFlags::NONBLOCK,
        Protocol::from_raw(0),
    )?;

    Ok((
        Notify { registered },
        NotifyToken {
            free: Arc::new(free),
        },
    ))
}

impl PartialEq<usize> for Key {
    fn eq(&self, other: &usize) -> bool {
        &self.0 == other
    }
}
impl PartialEq<Key> for usize {
    fn eq(&self, other: &Key) -> bool {
        self == &other.0
    }
}

impl AsFd for Io {
    fn as_fd(&self) -> rustix::fd::BorrowedFd<'_> {
        self.inner.as_fd()
    }
}

impl PollManager {
    pub fn new() -> rustix::io::Result<Self> {
        let (notify, notify_token) = notify_pair()?;

        Ok(Self {
            id: 0,
            notify,
            notify_token,
            notify_key: None,
            io: HashMap::new(),
        })
    }

    pub fn register<A: AsFd + 'static>(&mut self, io: Arc<A>, interests: Readiness) -> Key {
        let key = self.id;
        self.id += 1;

        self.io.insert(
            key,
            Managed {
                io: Io { inner: io },
                interests,
            },
        );

        Key(key)
    }

    pub fn update_interests(&mut self, key: &Key, interests: Readiness) {
        if let Some(ref mut managed) = self.io.get_mut(&key.0) {
            managed.interests = interests;
        }
    }

    pub fn deregister(&mut self, key: Key) {
        self.io.remove(&key.0);
    }

    pub fn notifier(&self) -> NotifyToken {
        self.notify_token.clone()
    }

    pub fn poll(&mut self, timeout: Option<i32>) -> rustix::io::Result<Vec<(usize, Readiness)>> {
        let mut poller = Poller::new();
        if let Some(timeout) = timeout {
            poller.timeout(timeout);
        }
        let notify_key = poller.add(
            &self.notify,
            Readiness::hangup() | Readiness::error() | Readiness::read(),
        );
        self.notify_key = Some(notify_key);

        let mut mapping = HashMap::new();

        for (mkey, managed) in self.io.iter() {
            let key = poller.add(&managed.io, managed.interests);

            mapping.insert(key, *mkey);
        }

        let events = poller
            .poll()?
            .into_iter()
            .filter_map(|(key, readiness)| {
                if key == notify_key {
                    let mut buf = [0; 16];
                    loop {
                        if let Err(e) = rustix::io::read(&self.notify, &mut buf) {
                            if e == rustix::io::Error::AGAIN || e == rustix::io::Error::WOULDBLOCK {
                                break;
                            }
                        }
                    }
                    None
                } else {
                    let mkey = mapping.remove(&key)?;
                    Some((mkey, readiness))
                }
            })
            .collect();

        Ok(events)
    }
}

impl Readiness {
    pub fn empty() -> Self {
        Self::default()
    }

    pub fn all() -> Self {
        Readiness {
            read: true,
            write: true,
            hangup: true,
            error: true,
        }
    }

    pub fn read() -> Self {
        Self {
            read: true,
            ..Default::default()
        }
    }

    pub fn write() -> Self {
        Self {
            write: true,
            ..Default::default()
        }
    }

    pub fn hangup() -> Self {
        Self {
            hangup: true,
            ..Default::default()
        }
    }

    pub fn error() -> Self {
        Self {
            error: true,
            ..Default::default()
        }
    }

    pub fn is_read(&self) -> bool {
        self.read
    }

    pub fn is_write(&self) -> bool {
        self.write
    }

    pub fn is_hangup(&self) -> bool {
        self.hangup
    }

    pub fn is_error(&self) -> bool {
        self.error
    }
}

impl BitOr for Readiness {
    type Output = Self;

    fn bitor(self, rhs: Self) -> Self::Output {
        Self {
            read: self.read || rhs.read,
            write: self.write || rhs.write,
            hangup: self.hangup || rhs.hangup,
            error: self.error || rhs.error,
        }
    }
}

impl From<PollFlags> for Readiness {
    fn from(flags: PollFlags) -> Self {
        Self {
            read: flags.intersects(PollFlags::IN),
            write: flags.intersects(PollFlags::OUT),
            hangup: flags.intersects(PollFlags::HUP),
            error: flags.intersects(PollFlags::ERR),
        }
    }
}

impl From<Readiness> for PollFlags {
    fn from(r: Readiness) -> Self {
        let mut output = PollFlags::empty();

        if r.read {
            output |= PollFlags::IN;
        }
        if r.write {
            output |= PollFlags::OUT;
        }
        if r.hangup {
            output |= PollFlags::HUP;
        }
        if r.error {
            output |= PollFlags::ERR;
        }

        output
    }
}

impl<'a> Poller<'a> {
    pub fn new() -> Self {
        Default::default()
    }

    pub fn add<'b: 'a, A: AsFd>(&mut self, fd: &'b A, interests: Readiness) -> usize {
        let len = self.fds.len();
        self.flags.push(interests.into());
        self.fds.push(PollFd::new(fd, interests.into()));
        len
    }

    pub fn timeout(&mut self, timeout: i32) {
        self.timeout = Some(timeout);
    }

    pub fn poll(mut self) -> rustix::io::Result<Vec<(usize, Readiness)>> {
        let timeout = self.timeout.unwrap_or(-1);

        let n = rustix::io::poll(&mut self.fds, timeout)?;

        Ok(self
            .fds
            .into_iter()
            .zip(self.flags)
            .enumerate()
            .filter_map(|(key, (pollfd, flags))| {
                if pollfd.revents().intersects(flags) {
                    Some((key, (pollfd.revents().into())))
                } else {
                    None
                }
            })
            .take(n)
            .collect())
    }
}

impl NotifyToken {
    pub fn notify(self) -> rustix::io::Result<Self> {
        if let Err(e) = rustix::io::write(&*self.free, &[0]) {
            if e == rustix::io::Error::AGAIN || e == rustix::io::Error::WOULDBLOCK {
                return Ok(self);
            }

            return Err(e);
        }

        Ok(self)
    }
}

impl AsFd for Notify {
    fn as_fd(&self) -> rustix::fd::BorrowedFd<'_> {
        self.registered.as_fd()
    }
}
