use rustix::{
    fd::AsFd,
    io::OwnedFd,
    net::{AddressFamily, Protocol, SocketFlags, SocketType},
};
use std::net::{IpAddr, SocketAddrV4, SocketAddrV6};

pub struct TcpStream {
    inner: OwnedFd,
}

pub struct TcpStreamBuilder {
    ip: IpAddr,
    port: u16,
    inner: OwnedFd,
}

impl TcpStreamBuilder {
    pub fn try_finish(self) -> rustix::io::Result<Result<TcpStream, TcpStreamBuilder>> {
        let res = match self.ip {
            IpAddr::V4(v4) => {
                let sock_addr = SocketAddrV4::new(v4, self.port);

                rustix::net::connect_v4(&self.inner, &sock_addr)
            }
            IpAddr::V6(v6) => {
                let flowinfo = 0;
                let scope_id = 0;
                let sock_addr = SocketAddrV6::new(v6, self.port, flowinfo, scope_id);

                rustix::net::connect_v6(&self.inner, &sock_addr)
            }
        };

        match res {
            Ok(()) => Ok(Ok(TcpStream { inner: self.inner })),
            Err(e) if e == rustix::io::Error::INPROGRESS => Ok(Err(self)),
            Err(e) => Err(e),
        }
    }
}

impl TcpStream {
    pub fn connect(ip: IpAddr, port: u16) -> rustix::io::Result<TcpStreamBuilder> {
        match ip {
            IpAddr::V4(_) => {
                let sock = rustix::net::socket_with(
                    AddressFamily::INET,
                    SocketType::STREAM,
                    SocketFlags::NONBLOCK,
                    Protocol::TCP,
                )?;

                Ok(TcpStreamBuilder {
                    ip,
                    port,
                    inner: sock,
                })
            }
            IpAddr::V6(_) => {
                let sock = rustix::net::socket_with(
                    AddressFamily::INET6,
                    SocketType::STREAM,
                    SocketFlags::NONBLOCK,
                    Protocol::TCP,
                )?;

                Ok(TcpStreamBuilder {
                    ip,
                    port,
                    inner: sock,
                })
            }
        }
    }
}

impl AsFd for TcpStream {
    fn as_fd(&self) -> rustix::fd::BorrowedFd<'_> {
        self.inner.as_fd()
    }
}

impl AsFd for TcpStreamBuilder {
    fn as_fd(&self) -> rustix::fd::BorrowedFd<'_> {
        self.inner.as_fd()
    }
}
